// This code is for a Pi Pico hardware addon for Ruby.
// It exposes these functionalities to Ruby controller:
// * Rotary Encoder
// * Cancel Button
// * RC In (SBUS and IBUS)
//
// Licence: you can copy, edit, change or do whatever you wish with the code in this file

#include "hardware/spi.h"
#include "hardware/gpio.h"
#include "pico/binary_info.h"
#include "pico/multicore.h"
#include "pico/util/queue.h"
#include <Wire.h>
#include "i2c_protocols.h"

#define PIN_I2C0_SDA 2
#define PIN_I2C0_SCL 3

#define PIN_UART_TX 4
#define PIN_UART_RX 5 
 
#define PIN_ROTARY_SEL 11
#define PIN_ROTARY_CLK 12
#define PIN_ROTARY_DATA 13

#define PIN_BUTTON_CANCEL 14

#define UART_ID uart1
#define BAUD_RATE_IBUS 115200
#define BAUD_RATE_SBUS 100000 

#define MAX_BUFFER_SIZE 1024

#define MAX_RC_CHANNELS 24
#define SBUS_NUM_CH 16
#define SBUS_BYTE_HEADER 0x0F
#define SBUS_BYTE_FOOTER 0x00
#define SBUS_BYTE_FOOTER2 0x04
#define SBUS_CH17_MASK 0x01
#define SBUS_CH18_MASK 0x02
#define SBUS_LOST_FRAME_MASK 0x04
#define SBUS_FAILSAFE_MASK 0x08  

typedef unsigned int u32;
typedef unsigned short u16;
typedef unsigned char u8; 

u8  s_nFrameNumber = 0; 
u8  s_nRCSBUS_Is_Failsafe = 0;
int s_nMustSetupUARTRCIN = 0;
int s_nDidSetupUARTRCIN = 0;
long s_lLastRCInChRead = 0;

u16 s_nChannels[3][MAX_RC_CHANNELS]; // 3 buffers to cycle through, one is currently reading from, one is currently writing to;
u16 s_nChannelsOutput[MAX_RC_CHANNELS];
unsigned char s_uRCChannelsOut[24];
u16 s_nChannelValue = 0;
u16 s_nChannelValueIndex = 0;

volatile u8 s_nChannelsIndexRead = 0;
volatile u8 s_nChannelsIndexWrite = 0;

int s_nIBUS_State = 0;
u16 s_IBusCheckSum = 0xFFFF;
u16 s_IBusCheckSumLow = 0; 
u16  s_uIBusFrameBuffer[32];

int s_nSBUS_State = 0;
u8 s_SBUS_prev_byte = SBUS_BYTE_FOOTER;
u8 s_SBUS_cur_byte = SBUS_BYTE_FOOTER;
u8 s_uSBusFrameBuffer[32];  

const unsigned char s_crc_i2c_table[256] = {
0x00,0x31,0x62,0x53,0xC4,0xF5,0xA6,0x97,0xB9,0x88,0xDB,0xEA,0x7D,0x4C,0x1F,0x2E,
0x43,0x72,0x21,0x10,0x87,0xB6,0xE5,0xD4,0xFA,0xCB,0x98,0xA9,0x3E,0x0F,0x5C,0x6D,
0x86,0xB7,0xE4,0xD5,0x42,0x73,0x20,0x11,0x3F,0x0E,0x5D,0x6C,0xFB,0xCA,0x99,0xA8,
0xC5,0xF4,0xA7,0x96,0x01,0x30,0x63,0x52,0x7C,0x4D,0x1E,0x2F,0xB8,0x89,0xDA,0xEB,
0x3D,0x0C,0x5F,0x6E,0xF9,0xC8,0x9B,0xAA,0x84,0xB5,0xE6,0xD7,0x40,0x71,0x22,0x13,
0x7E,0x4F,0x1C,0x2D,0xBA,0x8B,0xD8,0xE9,0xC7,0xF6,0xA5,0x94,0x03,0x32,0x61,0x50,
0xBB,0x8A,0xD9,0xE8,0x7F,0x4E,0x1D,0x2C,0x02,0x33,0x60,0x51,0xC6,0xF7,0xA4,0x95,
0xF8,0xC9,0x9A,0xAB,0x3C,0x0D,0x5E,0x6F,0x41,0x70,0x23,0x12,0x85,0xB4,0xE7,0xD6,
0x7A,0x4B,0x18,0x29,0xBE,0x8F,0xDC,0xED,0xC3,0xF2,0xA1,0x90,0x07,0x36,0x65,0x54,
0x39,0x08,0x5B,0x6A,0xFD,0xCC,0x9F,0xAE,0x80,0xB1,0xE2,0xD3,0x44,0x75,0x26,0x17,
0xFC,0xCD,0x9E,0xAF,0x38,0x09,0x5A,0x6B,0x45,0x74,0x27,0x16,0x81,0xB0,0xE3,0xD2,
0xBF,0x8E,0xDD,0xEC,0x7B,0x4A,0x19,0x28,0x06,0x37,0x64,0x55,0xC2,0xF3,0xA0,0x91,
0x47,0x76,0x25,0x14,0x83,0xB2,0xE1,0xD0,0xFE,0xCF,0x9C,0xAD,0x3A,0x0B,0x58,0x69,
0x04,0x35,0x66,0x57,0xC0,0xF1,0xA2,0x93,0xBD,0x8C,0xDF,0xEE,0x79,0x48,0x1B,0x2A,
0xC1,0xF0,0xA3,0x92,0x05,0x34,0x67,0x56,0x78,0x49,0x1A,0x2B,0xBC,0x8D,0xDE,0xEF,
0x82,0xB3,0xE0,0xD1,0x46,0x77,0x24,0x15,0x3B,0x0A,0x59,0x68,0xFF,0xCE,0x9D,0xAC };

unsigned char s_uI2CLastCommandBuff[MAX_BUFFER_SIZE];
unsigned int s_iI2CLastCommandExpectedLength = 0;
unsigned int s_iI2CLastCommandReadPosition = 0;
long s_lLastI2CCommandInTime = 0;

unsigned char s_uI2CLastResponseBuff[MAX_BUFFER_SIZE];
unsigned int s_iI2CLastResponseLength = 0;
unsigned int s_iI2CLastResponseSentPosition = -1;

long timeNow = 0;
long lastLedTime = 0;
int ledState = 0;

unsigned char s_uRCFlags = 0;

long lastReadRotaryEncoderTime = 0;
long lastReadRotaryEncoderButtonTime = 0;
int lastRotaryReadClk = 0;
int lastRotaryReadData = 0;
int lastRotaryReadSelect = 1; // pullup, 1 is not pressed
int lastRotaryPinChange = -1; // -1 none, 0 - clock, 1 - data
long lastRotartEncoderPressTime = 0;

int stateRotaryWasPressed = 0;
int stateRotaryWasLongPressed = 0;
int stateRotaryWasRotatedCWCount = 0;
int stateRotaryWasRotatedCCWCount = 0;

long lastReadButtonsTime = 0;
long lastButtonCancelPressTime = 0;
int lastButtonCancelState = 1; // pullup, 1 is not pressed
int stateButtonCancelWasPressed = 0;
int stateButtonCancelWasLongPressed = 0;

int countMsg = 0;
int countGlobal = 1000;
queue_t queue_to_core2;

void _dbg(const char* szText)
{
   //Serial.print(szText);
}

void _dbg_ln(const char* szText)
{
   //Serial.println(szText);
}

void _dbg(int i)
{
   //Serial.print(i);
}

void _dbg_ln(int i)
{
   //Serial.println(i);
}

/*
void core2_entry()
{

   _dbg_ln("Core 2 started");
   _dbg_ln("Core 2 waiting");
   while (1)
   {
      //delay(1900);
      //Serial.println("*");
      int c = 0;
      queue_remove_blocking(&queue_to_core2, &c);
      _dbg("Core 2 received: ");
      _dbg(c);
      _dbg_ln(" from Core 1");
      _dbg("Core 2 global counter is: ");
      _dbg_ln(countGlobal);
   }
}
*/

// The setup function runs once when you press reset or power the board

void setup()
{
   // Initialize digital pin LED_BUILTIN as an output.
   pinMode(LED_BUILTIN, OUTPUT);
   pinMode(PIN_ROTARY_CLK, INPUT_PULLUP);
   pinMode(PIN_ROTARY_DATA, INPUT_PULLUP);
   pinMode(PIN_ROTARY_SEL, INPUT_PULLUP);
   pinMode(PIN_BUTTON_CANCEL, INPUT_PULLUP);
  
   //Serial.begin (57600);
  
   Wire1.setSDA(PIN_I2C0_SDA);
   Wire1.setSCL(PIN_I2C0_SCL);
   Wire1.begin(I2C_DEVICE_MAX_ADDRESS_RANGE-2);
   Wire1.onReceive(onI2C0Recv);
   Wire1.onRequest(onI2C0Req);

   for( int i=0; i<24; i++ )
      s_uRCChannelsOut[i] = 0;
     
   lastLedTime = millis();

   lastRotaryReadClk = digitalRead(PIN_ROTARY_CLK);
   lastRotaryReadData = digitalRead(PIN_ROTARY_DATA);
   lastRotaryReadSelect = digitalRead(PIN_ROTARY_SEL);
   lastButtonCancelState = digitalRead(PIN_BUTTON_CANCEL);
  
   //queue_init(&queue_to_core2, sizeof(int), 2000);
   //_dbg_ln("Starting second core...");
   //multicore_launch_core1(core2_entry);
   //_dbg_ln("Multicore launch done.");
}

// the loop function runs over and over again forever
void loop() 
{
   timeNow = millis();

   if ( s_nMustSetupUARTRCIN )
   {
     _init_rc_in_parse_state();
     uart_deinit(UART_ID);
     _setup_uart();
     s_nMustSetupUARTRCIN = 0;
   }
   _read_rotary_encoder();
   _read_buttons();
   _alive();

   if ( s_nDidSetupUARTRCIN == 1 )
   {
      long lCountRead = 0;
      while ( uart_is_readable(UART_ID) > 0 && lCountRead < 50 )
      {
         u8 ch = uart_getc(UART_ID);
         timeNow = millis();
         if ( timeNow > s_lLastRCInChRead + 3 )
         {
            s_nIBUS_State = 0;
            s_nSBUS_State = 0;
            s_SBUS_prev_byte = SBUS_BYTE_FOOTER;
            s_SBUS_cur_byte = SBUS_BYTE_FOOTER;
         }
         s_lLastRCInChRead = timeNow;
         
         if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_IBUS )
            _parse_serial_byte_ibus(ch);
         else if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_SBUS )
            _parse_serial_byte_sbus(ch);
         lCountRead++;
      }
   }
}

void _alive()
{
  if ( s_nFrameNumber != 0 )
  {
     if ( timeNow < lastLedTime + 100 )
        return;    
  }
  else
     if ( timeNow < lastLedTime + 990 )
        return;
     
  lastLedTime = timeNow;
  ledState = 1 - ledState;
  if ( ledState )
  {
     digitalWrite(LED_BUILTIN, HIGH);   // turn the LED on (HIGH is the voltage level)
     countGlobal++;
     if ( (countGlobal % 20) == 0 )
     {
        countMsg++;
     }
  }
  else
     digitalWrite(LED_BUILTIN, LOW);    // turn the LED off by making the voltage LOW
}

void _setup_uart()
{
   s_nDidSetupUARTRCIN = 0;
   if (  (!( s_uRCFlags & I2C_COMMAND_RC_FLAG_IBUS )) && (!( s_uRCFlags & I2C_COMMAND_RC_FLAG_SBUS )) )
      return;

   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_INVERT_UART )
      gpio_set_inover(PIN_UART_RX, GPIO_OVERRIDE_INVERT);
   else
      gpio_set_inover(PIN_UART_RX, GPIO_OVERRIDE_NORMAL);

   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_IBUS )
      uart_init(UART_ID, BAUD_RATE_IBUS);
   else if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_SBUS )
      uart_init(UART_ID, BAUD_RATE_SBUS);

   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_INVERT_UART )
      gpio_set_inover(PIN_UART_RX, GPIO_OVERRIDE_INVERT);
   else
      gpio_set_inover(PIN_UART_RX, GPIO_OVERRIDE_NORMAL);

   gpio_set_function(PIN_UART_TX, GPIO_FUNC_UART);
   gpio_set_function(PIN_UART_RX, GPIO_FUNC_UART);

   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_INVERT_UART )
      gpio_set_inover(PIN_UART_RX, GPIO_OVERRIDE_INVERT);
   else
      gpio_set_inover(PIN_UART_RX, GPIO_OVERRIDE_NORMAL);

   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_IBUS )
   {
      uart_set_baudrate(UART_ID, BAUD_RATE_IBUS);
      uart_set_format(UART_ID, 8,1, UART_PARITY_NONE);
   }
   else if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_SBUS )
   {
      uart_set_baudrate(UART_ID, BAUD_RATE_SBUS);
      uart_set_format(UART_ID, 8,2, UART_PARITY_EVEN);
   }

   uart_set_hw_flow(UART_ID, false, false);

   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_INVERT_UART )
      gpio_set_inover(PIN_UART_RX, GPIO_OVERRIDE_INVERT);
   else
      gpio_set_inover(PIN_UART_RX, GPIO_OVERRIDE_NORMAL);

   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_SBUS )
      _dbg_ln("Did setup UART for RC In for SBUS");
   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_IBUS )
      _dbg_ln("Did setup UART for RC In for IBUS");
      
   if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_INVERT_UART )
      _dbg_ln("Did setup UART for RC In Inverted.");
   else
      _dbg_ln("Did setup UART for RC In Uninverted.");
   s_nDidSetupUARTRCIN = 1;
   s_lLastRCInChRead = millis();
}

void _init_rc_in_parse_state()
{
   s_nIBUS_State = 0;
   s_nSBUS_State = 0;
   s_SBUS_prev_byte = SBUS_BYTE_FOOTER;
   s_SBUS_cur_byte = SBUS_BYTE_FOOTER;
   s_nRCSBUS_Is_Failsafe = 0;
   
   for( int i=0; i<MAX_RC_CHANNELS; i++ )
   {
      s_nChannels[0][i] = 0;
      s_nChannels[1][i] = 0;
      s_nChannels[2][i] = 0;
   }

   s_nChannelValue = 0;
   s_nChannelValueIndex = 0;

   s_nChannelsIndexRead = 0;
   s_nChannelsIndexWrite = 1;
   s_nFrameNumber = 0; 

   for( int i=0; i<24; i++ )
     s_uRCChannelsOut[i] = 0;
}

static int _parse_serial_byte_ibus(u8 value)
{
   switch(s_nIBUS_State)
   {
      case 0:
        if (value == 0x20)
        {
           s_IBusCheckSum = 0xFFFF - value;
           s_nIBUS_State = 1;
           return 0;
        }
        return 0;

      case 1:
        if (value == 0x40)
        {
           s_IBusCheckSum -= value;
           s_nIBUS_State = 2;
           return 0;
        }
        else
           s_nIBUS_State = 0;
        return 0;

     case 30:
        {
           s_IBusCheckSumLow = value;
           s_nIBUS_State = 31;
           return 0;
        }
      
     case 31:
        {
          s_nIBUS_State = 0;
          u16 checksum = (value << 8) | s_IBusCheckSumLow;
          if (checksum == s_IBusCheckSum)
          {
             on_new_ibus_frame();
             return 1;
          }
        }
        return 0;

      default:
      {
         if ( s_nIBUS_State >= 2 && s_nIBUS_State < 31 )
            s_uIBusFrameBuffer[s_nIBUS_State-2] = value;
         s_IBusCheckSum -= value;
         s_nIBUS_State++;
         if ( s_nIBUS_State > 31 )
         {
            s_nIBUS_State = 0;
            return 0;
         }
      }
   }
   return 0;
}

int _parse_serial_byte_sbus(u8 value)
{
   s_SBUS_prev_byte = s_SBUS_cur_byte;
   s_SBUS_cur_byte = value;

   if ( s_nSBUS_State == 0 )
   {
      if ((s_SBUS_cur_byte == SBUS_BYTE_HEADER) && ((s_SBUS_prev_byte == SBUS_BYTE_FOOTER) ||
         ((s_SBUS_prev_byte & 0x0F) == SBUS_BYTE_FOOTER2)))
      {
        s_uSBusFrameBuffer[0] = s_SBUS_cur_byte;
        s_nSBUS_State = 1;
      }
      return 0;
    }

    if ( s_nSBUS_State < 24 )
    {
       s_uSBusFrameBuffer[s_nSBUS_State] = s_SBUS_cur_byte;
       s_nSBUS_State++;
       return 0;
    }

    s_nSBUS_State = 0;
    if ((s_SBUS_cur_byte == SBUS_BYTE_FOOTER) || ((s_SBUS_cur_byte & 0x0F) == SBUS_BYTE_FOOTER2))
    {
      on_new_sbus_frame();
      return 1;
    }
    return 0;
}

static void on_new_ibus_frame()
{
   _dbg_ln(s_nChannels[s_nChannelsIndexWrite][0]);
   for( int i=0; i<14; i++ )
      s_nChannels[s_nChannelsIndexWrite][i] = s_uIBusFrameBuffer[i*2] + (s_uIBusFrameBuffer[i*2+1]<<8);

   for( int i=0; i<16; i++ )
      s_uRCChannelsOut[i] = s_nChannels[s_nChannelsIndexWrite][i] & 0xFF;
   for( int i=0; i<8; i++ )
   {
      s_uRCChannelsOut[i+16] = (s_nChannels[s_nChannelsIndexWrite][i*2] >> 8) & 0x0F;
      s_uRCChannelsOut[i+16] |= (((s_nChannels[s_nChannelsIndexWrite][i*2+1] >> 8) & 0x0F)<<4);
   }

   s_nChannelsIndexWrite++;
   if ( s_nChannelsIndexWrite > 2 )
      s_nChannelsIndexWrite = 0;
   s_nChannelsIndexRead++;
   
   if ( s_nChannelsIndexRead > 2 )
      s_nChannelsIndexRead = 0;
   s_nFrameNumber++;
}

static void on_new_sbus_frame()
{
   u16 ch_[24];

   ch_[0]  = (u16)(((s_uSBusFrameBuffer[1]     ) | (s_uSBusFrameBuffer[2] << 8)) & 0x07FF);
   ch_[1]  = (u16)(((s_uSBusFrameBuffer[2] >> 3) | (s_uSBusFrameBuffer[3] << 5)) & 0x07FF);
   ch_[2]  = (u16)(((s_uSBusFrameBuffer[3] >> 6) | (s_uSBusFrameBuffer[4] << 2)  | (s_uSBusFrameBuffer[5] << 10)) & 0x07FF);
   ch_[3]  = (u16)(((s_uSBusFrameBuffer[5] >> 1) | (s_uSBusFrameBuffer[6] << 7)) & 0x07FF);
   ch_[4]  = (u16)(((s_uSBusFrameBuffer[6] >> 4) | (s_uSBusFrameBuffer[7] << 4)) & 0x07FF);
   ch_[5]  = (u16)(((s_uSBusFrameBuffer[7] >> 7) | (s_uSBusFrameBuffer[8] << 1)  | (s_uSBusFrameBuffer[9] << 9)) & 0x07FF);
   ch_[6]  = (u16)(((s_uSBusFrameBuffer[9] >> 2) | (s_uSBusFrameBuffer[10]<< 6)) & 0x07FF);
   ch_[7]  = (u16)(((s_uSBusFrameBuffer[10]>> 5) | (s_uSBusFrameBuffer[11]<< 3)) & 0x07FF);
   ch_[8]  = (u16)(((s_uSBusFrameBuffer[12]    ) | (s_uSBusFrameBuffer[13]<< 8)) & 0x07FF);
   ch_[9]  = (u16)(((s_uSBusFrameBuffer[13]>> 3) | (s_uSBusFrameBuffer[14]<< 5)) & 0x07FF);
   ch_[10] = (u16)(((s_uSBusFrameBuffer[14]>> 6) | (s_uSBusFrameBuffer[15]<< 2)  | (s_uSBusFrameBuffer[16] << 10)) & 0x07FF);
   ch_[11] = (u16)(((s_uSBusFrameBuffer[16]>> 1) | (s_uSBusFrameBuffer[17]<< 7)) & 0x07FF);
   ch_[12] = (u16)(((s_uSBusFrameBuffer[17]>> 4) | (s_uSBusFrameBuffer[18]<< 4)) & 0x07FF);
   ch_[13] = (u16)(((s_uSBusFrameBuffer[18]>> 7) | (s_uSBusFrameBuffer[19]<< 1)  | (s_uSBusFrameBuffer[20] << 9)) & 0x07FF);
   ch_[14] = (u16)(((s_uSBusFrameBuffer[20]>> 2) | (s_uSBusFrameBuffer[21]<< 6)) & 0x07FF);
   ch_[15] = (u16)(((s_uSBusFrameBuffer[21]>> 5) | (s_uSBusFrameBuffer[22]<< 3)) & 0x07FF);
   
   // CH 17
   ch_[16] = (s_uSBusFrameBuffer[23] & SBUS_CH17_MASK)?1:0;
   // CH 18
   ch_[17] = (s_uSBusFrameBuffer[23] & SBUS_CH18_MASK)?1:0;
    
   // Get the lost frame flag
   int lost_frame = (s_uSBusFrameBuffer[23] & SBUS_LOST_FRAME_MASK)?1:0;
   // Get the failsafe flag
   s_nRCSBUS_Is_Failsafe = (s_uSBusFrameBuffer[23] & SBUS_FAILSAFE_MASK)?1:0;

   if ( s_nRCSBUS_Is_Failsafe )
   {
      return;
   }

   for( int i=0; i<18; i++ )
      s_nChannels[s_nChannelsIndexWrite][i] = ch_[i];

   for( int i=0; i<16; i++ )
      s_uRCChannelsOut[i] = s_nChannels[s_nChannelsIndexWrite][i] & 0xFF;
   for( int i=0; i<8; i++ )
   {
      s_uRCChannelsOut[i+16] = (s_nChannels[s_nChannelsIndexWrite][i*2] >> 8) & 0x0F;
      s_uRCChannelsOut[i+16] |= (((s_nChannels[s_nChannelsIndexWrite][i*2+1] >> 8) & 0x0F)<<4);
   }

   s_nChannelsIndexWrite++;
   if ( s_nChannelsIndexWrite > 2 )
      s_nChannelsIndexWrite = 0;
   s_nChannelsIndexRead++;
   if ( s_nChannelsIndexRead > 2 )
      s_nChannelsIndexRead = 0;
   s_nFrameNumber++;
} 

void _read_rotary_encoder()
{
  if ( timeNow < lastReadRotaryEncoderTime + 2 )
     return;
  lastReadRotaryEncoderTime = timeNow;

  if ( timeNow > lastReadRotaryEncoderButtonTime + 10 )
  {
    lastReadRotaryEncoderButtonTime = timeNow;
     int readSelect = digitalRead(PIN_ROTARY_SEL);
     if ( readSelect != lastRotaryReadSelect )
     {
        if ( 0 == readSelect )
           lastRotartEncoderPressTime = timeNow;
        else if ( 0 != lastRotartEncoderPressTime )
        {
         if ( timeNow >= lastRotartEncoderPressTime + 1000 )
         {
            stateRotaryWasLongPressed = 1;
            _dbg_ln("Event: Rotary long press");
         }
         else
         {
            stateRotaryWasPressed = 1;
            _dbg_ln("Event: Rotary press");
         }
         lastRotartEncoderPressTime = 0;
        }
        lastRotaryReadSelect = readSelect;
     }
     else if ( 0 == readSelect && ( 0 != lastRotartEncoderPressTime ) )
     {
       if ( timeNow >= lastRotartEncoderPressTime + 1000 )
         {
            lastRotartEncoderPressTime = 0;
            stateRotaryWasLongPressed = 1;
            _dbg_ln("Event: Rotary long press");
         }
     }
  }
  
  int readClk = digitalRead(PIN_ROTARY_CLK);
  readClk = digitalRead(PIN_ROTARY_CLK);
  int readData = digitalRead(PIN_ROTARY_DATA);
  readData = digitalRead(PIN_ROTARY_DATA);

  if ( readClk != lastRotaryReadClk )
  if ( readData != lastRotaryReadData )
     return;

  if ( readClk != lastRotaryReadClk )
  {
    if ( lastRotaryPinChange == 1 )
    {
       if ( readClk == 0 && readData == 1 )
       {
          _dbg("+");
          stateRotaryWasRotatedCWCount++;
       }
       if ( readClk == 1 && readData == 0 )
       {
          _dbg("+");
          stateRotaryWasRotatedCWCount++;
       }
       if ( readClk == 0 && readData == 0 )
       {
          _dbg("-");
          stateRotaryWasRotatedCCWCount++;
       }
       if ( readClk == 1 && readData == 1 )
       {
          _dbg("-");
          stateRotaryWasRotatedCCWCount++;
       }
    }
    lastRotaryPinChange = 0;
  }
  if ( readData != lastRotaryReadData )
  {
    lastRotaryPinChange = 1;
  }
  lastRotaryReadClk = readClk;
  lastRotaryReadData = readData;
}

void _read_buttons()
{
  if ( timeNow < lastReadButtonsTime + 10 )
     return;
  lastReadButtonsTime = timeNow;
  
  int iCancelButton = digitalRead(PIN_BUTTON_CANCEL);
  if ( iCancelButton != lastButtonCancelState )
  {
     if ( 0 == iCancelButton )
        lastButtonCancelPressTime = timeNow;
     else if ( 0 != lastButtonCancelPressTime )
     {
         if ( timeNow >= lastButtonCancelPressTime + 1000 )
         {
            stateButtonCancelWasLongPressed = 1;
            _dbg_ln("Event: Button cancel long press");
         }
         else
         {
            stateButtonCancelWasPressed = 1;
            _dbg_ln("Event: Button cancel press");
         }
         lastButtonCancelPressTime = 0;
     }
     lastButtonCancelState = iCancelButton;
  }
  else if ( 0 == iCancelButton && ( 0 != lastButtonCancelPressTime ) )
  {
    if ( timeNow >= lastButtonCancelPressTime + 1000 )
         {
            stateButtonCancelWasLongPressed = 1;
            lastButtonCancelPressTime = 0;
            _dbg_ln("Event: Button cancel long press");
         }
  }
}

void _parse_I2C_command()
{
  if ( s_iI2CLastCommandReadPosition < 3 )
     return;
     
  unsigned char crc = 0xFF;
  for( int i=0; i<s_iI2CLastCommandReadPosition-1; i++ )
     crc = s_crc_i2c_table[s_uI2CLastCommandBuff[i] ^ crc];
     
  if ( crc != s_uI2CLastCommandBuff[s_iI2CLastCommandReadPosition-1] )
  {
     _dbg_ln("Invalid command CRC");
     return;
  }

  s_iI2CLastResponseLength = 0;
  s_iI2CLastResponseSentPosition = -1;
  
  if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_FLAGS )
  {
     _dbg_ln("Get flags command");
     s_uI2CLastResponseBuff[0] = I2C_CAPABILITY_FLAG_ROTARY | I2C_CAPABILITY_FLAG_BUTTONS | I2C_CAPABILITY_FLAG_RC_INPUT;
     s_uI2CLastResponseBuff[1] = 0;
     s_iI2CLastResponseLength = 2;
     s_iI2CLastResponseSentPosition = -1;
  }
  if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_NAME )
  {
     _dbg_ln("Get name command");
     strcpy((char*)s_uI2CLastResponseBuff, "Pico Ruby New Extender");
     s_iI2CLastResponseLength = I2C_PROTOCOL_STRING_LENGTH;
     s_iI2CLastResponseSentPosition = -1;
  }

  if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_ROTARY_EVENTS )
  {
     //_dbg_ln("Get rotary events command");
     s_uI2CLastResponseBuff[0] = 0;

     if ( stateRotaryWasPressed )
        s_uI2CLastResponseBuff[0] |= 1;
     if ( stateRotaryWasLongPressed )
        s_uI2CLastResponseBuff[0] |= (1<<1);

     if ( stateRotaryWasRotatedCWCount > stateRotaryWasRotatedCCWCount )
     {
        s_uI2CLastResponseBuff[0] |= (1<<3);
        if ( stateRotaryWasRotatedCWCount/2 >= 3 )
           s_uI2CLastResponseBuff[0] |= (1<<5);        
     }
     else if ( stateRotaryWasRotatedCWCount < stateRotaryWasRotatedCCWCount )
     {
        s_uI2CLastResponseBuff[0] |= (1<<2);
        if ( stateRotaryWasRotatedCCWCount/2 >= 3 )
           s_uI2CLastResponseBuff[0] |= (1<<4);
     }
          
     s_iI2CLastResponseLength = 1;
     s_iI2CLastResponseSentPosition = -1;

     stateRotaryWasPressed = 0;
     stateRotaryWasLongPressed = 0;
     stateRotaryWasRotatedCWCount = 0;
     stateRotaryWasRotatedCCWCount = 0;
  }

  if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_BUTTONS_EVENTS )
  {
     //_dbg_ln("Get buttons events command");
     s_uI2CLastResponseBuff[0] = 0;
     s_uI2CLastResponseBuff[1] = 0;
     s_uI2CLastResponseBuff[2] = 0;
     s_uI2CLastResponseBuff[3] = 0;

     if ( stateButtonCancelWasPressed )
        s_uI2CLastResponseBuff[0] |= (1<<1);
     if ( stateButtonCancelWasLongPressed )
        s_uI2CLastResponseBuff[2] |= (1<<1);

     s_iI2CLastResponseLength = 4;
     s_iI2CLastResponseSentPosition = -1;

     stateButtonCancelWasPressed = 0;
     stateButtonCancelWasLongPressed = 0;
  }

  if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_SET_RC_INPUT_FLAGS )
  {
     s_uRCFlags = s_uI2CLastCommandBuff[2];
     _dbg("Set RC flags command. Protocol: ");
     if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_SBUS )
        _dbg("SBUS. Inverted: ");
     if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_IBUS )
        _dbg("IBUS. Inverted: ");
     if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_PPM )
        _dbg("PPM. Inverted: ");

     if ( s_uRCFlags & I2C_COMMAND_RC_FLAG_INVERT_UART )
        _dbg_ln("Yes");
     else
        _dbg_ln("No");
     s_uI2CLastResponseBuff[0] = 0;
     s_iI2CLastResponseLength = 1;
     s_iI2CLastResponseSentPosition = -1;

     s_nMustSetupUARTRCIN = 1;
  }
  
  if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_RC_GET_CHANNELS )
  {
     s_uI2CLastResponseBuff[0] = s_nRCSBUS_Is_Failsafe;
     s_uI2CLastResponseBuff[1] = s_nFrameNumber;
     for( int i=0; i<24; i++ )
        s_uI2CLastResponseBuff[i+2] = s_uRCChannelsOut[i];
     s_iI2CLastResponseLength = 26;
     s_iI2CLastResponseSentPosition = -1;
  }
  
  if ( s_iI2CLastResponseLength > 0 )
  {
    unsigned char crc = 0xFF;
    for( int i=0; i<s_iI2CLastResponseLength; i++ )
       crc = s_crc_i2c_table[s_uI2CLastResponseBuff[i] ^ crc];
       
    s_uI2CLastResponseBuff[s_iI2CLastResponseLength] = crc;
    s_iI2CLastResponseLength++;
  }
}


void onI2C0Recv(int len)
{
  s_lLastI2CCommandInTime = timeNow;
  
  for (int i = 0; i < len; i++)
  {
    int ch = Wire1.read();
    if ( ch == I2C_COMMAND_START_FLAG )
       s_iI2CLastCommandReadPosition = 0;
    s_uI2CLastCommandBuff[s_iI2CLastCommandReadPosition] = ch;
    s_iI2CLastCommandReadPosition++;
    
    if ( 2 == s_iI2CLastCommandReadPosition )
    {
      s_iI2CLastCommandExpectedLength = 0;
      if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_FLAGS )
         s_iI2CLastCommandExpectedLength = 1;
      if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_NAME )
         s_iI2CLastCommandExpectedLength = 1;
      if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_BUTTONS_EVENTS )
         s_iI2CLastCommandExpectedLength = 1;
      if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_ROTARY_EVENTS )
         s_iI2CLastCommandExpectedLength = 1;
      if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_SET_RC_INPUT_FLAGS )
         s_iI2CLastCommandExpectedLength = 2;
      if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_RC_GET_CHANNELS )
         s_iI2CLastCommandExpectedLength = 1;

      // If we don't know about the command, just ignore it and reset the buffer
      if ( s_iI2CLastCommandExpectedLength == 0 )
         s_iI2CLastCommandReadPosition = 0;
      else
         s_iI2CLastCommandExpectedLength+=2; // add 1 byte for start command byte and 1 byte for CRC
    }
    
    if ( s_iI2CLastCommandReadPosition == s_iI2CLastCommandExpectedLength )
    {
       _parse_I2C_command();
       s_iI2CLastCommandReadPosition = 0;
       s_iI2CLastCommandExpectedLength = 0;
    }
    if ( s_iI2CLastCommandReadPosition >= MAX_BUFFER_SIZE )
    {
       s_iI2CLastCommandReadPosition = 0;
       s_iI2CLastCommandExpectedLength = 0;
    }
  }
}

// Called when the I2C slave is read from
void onI2C0Req()
{
  s_iI2CLastResponseSentPosition++;
  if ( s_iI2CLastResponseSentPosition >= MAX_BUFFER_SIZE || s_iI2CLastResponseSentPosition >= s_iI2CLastResponseLength)
  {
     _dbg_ln("Error: unexpected request from master!");
     s_iI2CLastResponseSentPosition = MAX_BUFFER_SIZE;
     s_iI2CLastCommandReadPosition = 0;
     s_uI2CLastCommandBuff[1] = 0;
     Wire1.write(0);
     return;
  }
  if ( s_uI2CLastCommandBuff[1] == I2C_COMMAND_ID_GET_FLAGS )
  {
     _dbg("Send flags back, byte ");
     _dbg(s_iI2CLastResponseSentPosition+1);
     _dbg(": ");
     _dbg_ln((int)s_uI2CLastResponseBuff[s_iI2CLastResponseSentPosition]);
  }
  
  Wire1.write(&s_uI2CLastResponseBuff[s_iI2CLastResponseSentPosition], 1);
}
